#=========================================================================
# StructuralRTLIRGenL2Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 3, 2019
"""Provide L2 structural RTLIR generation pass."""

from .StructuralRTLIRGenL1Pass import StructuralRTLIRGenL1Pass


class StructuralRTLIRGenL2Pass( StructuralRTLIRGenL1Pass ):
  pass
