#=========================================================================
# StructuralRTLIRGenL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 3, 2019
"""Provide L3 structural RTLIR generation pass."""

from .StructuralRTLIRGenL2Pass import StructuralRTLIRGenL2Pass
from .StructuralRTLIRSignalExpr import CurComp


class StructuralRTLIRGenL3Pass( StructuralRTLIRGenL2Pass ):
  pass
