#=========================================================================
# SVBehavioralTranslatorL0.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 0 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL0 import (
    BehavioralTranslatorL0,
)


class SVBehavioralTranslatorL0( BehavioralTranslatorL0 ):
  pass
